package jp.hishidama.javadb.tool.data;

import java.math.BigDecimal;
import java.sql.Time;
import java.sql.Timestamp;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.Map;

import jp.hishidama.javadb.tool.ToolProperties;

public class Data implements Comparable<Data>, Cloneable {

	protected Class<?> clazz;

	protected Converter conv;

	protected Object value;

	protected boolean keyword;

	public Data(Class<?> cls) {
		this.clazz = cls;
		for (;;) {
			Converter c = map.get(cls);
			if (c != null) {
				conv = c;
				break;
			}
			cls = cls.getSuperclass();
			if (cls == null) {
				conv = map.get(Object.class);
				break;
			}
		}
		if (conv == null) {
			throw new IllegalArgumentException("class=" + clazz);
		}
		value = "null";
		keyword = true;
	}

	public void setValue(Object value) {
		this.value = value;
		keyword = false;
	}

	public Object getValue() {
		return value;
	}

	public void setKeyword(boolean b) {
		keyword = b;
	}

	public boolean isKeyword() {
		return keyword;
	}

	public void setValue(String str, boolean k) throws Exception {
		if (str == null) {
			value = null;
		} else {
			if (k) {
				value = str;
			} else {
				value = conv.valueOf(str);
			}
		}
		keyword = k;
	}

	@Override
	public String toString() {
		if (value == null) {
			return "";
		}
		if (keyword) {
			return (String) value;
		} else {
			return conv.toString(value);
		}
	}

	public Class<?> getValueClass() {
		return clazz;
	}

	@Override
	public boolean equals(Object obj) {
		if (obj instanceof Data) {
			return compareTo((Data) obj) == 0;
		}
		return false;
	}

	@Override
	public int hashCode() {
		int hash = clazz.hashCode();
		if (value != null) {
			hash ^= value.hashCode();
		}
		if (keyword) {
			hash = ~hash;
		}
		return hash;
	}

	@Override
	public int compareTo(Data o) {
		if (this.keyword && o.keyword) {
			return this.toString().compareTo(o.toString());
		}
		if (!this.keyword && o.keyword) {
			return -1;
		}
		if (this.keyword && !o.keyword) {
			return +1;
		}
		Object v1 = this.value;
		Object v2 = o.value;
		if (v1 == null) {
			if (v2 == null) {
				return 0;
			} else {
				return -1;
			}
		} else {
			if (v2 == null) {
				return +1;
			} else {
				return conv.compare(v1, v2);
			}
		}
	}

	static interface Converter {
		public Object valueOf(String str) throws Exception;

		public String toString(Object obj);

		public int compare(Object value1, Object value2);
	}

	static Map<Class<?>, Converter> map = new LinkedHashMap<Class<?>, Converter>();

	static {
		map.put(String.class, new Converter() {
			@Override
			public String toString(Object obj) {
				return (String) obj;
			}

			@Override
			public Object valueOf(String str) {
				return str;
			}

			@Override
			public int compare(Object value1, Object value2) {
				String v1 = (String) value1;
				String v2 = (String) value2;
				return v1.compareTo(v2);
			}
		});

		map.put(Byte.class, new Converter() {
			@Override
			public String toString(Object obj) {
				Byte i = (Byte) obj;
				return i.toString();
			}

			@Override
			public Object valueOf(String str) {
				return Byte.valueOf(str);
			}

			@Override
			public int compare(Object value1, Object value2) {
				Byte v1 = (Byte) value1;
				Byte v2 = (Byte) value2;
				return v1.compareTo(v2);
			}
		});
		map.put(Short.class, new Converter() {
			@Override
			public String toString(Object obj) {
				Short i = (Short) obj;
				return i.toString();
			}

			@Override
			public Object valueOf(String str) {
				return Short.valueOf(str);
			}

			@Override
			public int compare(Object value1, Object value2) {
				Short v1 = (Short) value1;
				Short v2 = (Short) value2;
				return v1.compareTo(v2);
			}
		});
		map.put(Integer.class, new Converter() {
			@Override
			public String toString(Object obj) {
				Integer i = (Integer) obj;
				return i.toString();
			}

			@Override
			public Object valueOf(String str) {
				return Integer.valueOf(str);
			}

			@Override
			public int compare(Object value1, Object value2) {
				Integer v1 = (Integer) value1;
				Integer v2 = (Integer) value2;
				return v1.compareTo(v2);
			}
		});
		map.put(Long.class, new Converter() {
			@Override
			public String toString(Object obj) {
				Long i = (Long) obj;
				return i.toString();
			}

			@Override
			public Object valueOf(String str) {
				return Long.valueOf(str);
			}

			@Override
			public int compare(Object value1, Object value2) {
				Long v1 = (Long) value1;
				Long v2 = (Long) value2;
				return v1.compareTo(v2);
			}
		});
		map.put(Float.class, new Converter() {
			@Override
			public String toString(Object obj) {
				Float i = (Float) obj;
				return i.toString();
			}

			@Override
			public Object valueOf(String str) {
				return Float.valueOf(str);
			}

			@Override
			public int compare(Object value1, Object value2) {
				Float v1 = (Float) value1;
				Float v2 = (Float) value2;
				return v1.compareTo(v2);
			}
		});
		map.put(Double.class, new Converter() {
			@Override
			public String toString(Object obj) {
				Double i = (Double) obj;
				return i.toString();
			}

			@Override
			public Object valueOf(String str) {
				return Double.valueOf(str);
			}

			@Override
			public int compare(Object value1, Object value2) {
				Double v1 = (Double) value1;
				Double v2 = (Double) value2;
				return v1.compareTo(v2);
			}
		});
		map.put(BigDecimal.class, new Converter() {
			@Override
			public String toString(Object obj) {
				BigDecimal i = (BigDecimal) obj;
				return i.toString();
			}

			@Override
			public Object valueOf(String str) {
				return new BigDecimal(str);
			}

			@Override
			public int compare(Object value1, Object value2) {
				BigDecimal v1 = (BigDecimal) value1;
				BigDecimal v2 = (BigDecimal) value2;
				return v1.compareTo(v2);
			}
		});

		abstract class DateConveter implements Converter {

			abstract SimpleDateFormat getFormat();

			@Override
			public String toString(Object obj) {
				SimpleDateFormat f = getFormat();
				Date d = (Date) obj;
				return f.format(d);
			}

			@Override
			public Object valueOf(String str) throws Exception {
				SimpleDateFormat f = getFormat();
				Date d = f.parse(str);
				return create(d.getTime());
			}

			protected abstract Object create(long date);

			@Override
			public int compare(Object value1, Object value2) {
				Date v1 = (Date) value1;
				Date v2 = (Date) value2;
				return v1.compareTo(v2);
			}
		}
		map.put(java.sql.Date.class, new DateConveter() {
			@Override
			SimpleDateFormat getFormat() {
				return ToolProperties.instance.getDateDF();
			}

			@Override
			protected Object create(long date) {
				return new java.sql.Date(date);
			}
		});
		map.put(Time.class, new DateConveter() {
			@Override
			SimpleDateFormat getFormat() {
				return ToolProperties.instance.getTimeDF();
			}

			@Override
			protected Object create(long date) {
				return new Time(date);
			}
		});
		map.put(Timestamp.class, new DateConveter() {
			@Override
			SimpleDateFormat getFormat() {
				return ToolProperties.instance.getTimestampDF();
			}

			@Override
			protected Object create(long date) {
				return new Timestamp(date);
			}
		});

		map.put(Boolean.class, new Converter() {
			@Override
			public String toString(Object obj) {
				Boolean b = (Boolean) obj;
				return b.toString();
			}

			@Override
			public Object valueOf(String str) throws Exception {
				return Boolean.valueOf(str);
			}

			@Override
			public int compare(Object value1, Object value2) {
				Boolean v1 = (Boolean) value1;
				Boolean v2 = (Boolean) value2;
				return v1.compareTo(v2);
			}
		});

		map.put(Object.class, new Converter() {
			@Override
			public String toString(Object obj) {
				return obj.toString();
			}

			@Override
			public Object valueOf(String str) throws Exception {
				throw new UnsupportedOperationException("valueOf(\"" + str
						+ "\")");
			}

			@Override
			public int compare(Object value1, Object value2) {
				throw new UnsupportedOperationException("compare(" + value1
						+ ", " + value2 + ")");
			}
		});
	}

	@Override
	protected Data clone() throws CloneNotSupportedException {
		return (Data) super.clone();
	}
}
