package jp.hishidama.javadb.tool.table;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

import jp.hishidama.javadb.tool.ToolUtil;
import jp.hishidama.javadb.tool.conn.ExecuteQuery;
import jp.hishidama.javadb.tool.conn.ConnManager;
import jp.hishidama.javadb.tool.data.Data;
import jp.hishidama.javadb.tool.schema.DerbySchemaFrame;
import jp.hishidama.swing.table.LazyTable;
import jp.hishidama.swing.table.LazyTableModel;

class TableDataModel extends LazyTableModel {
	private static final long serialVersionUID = -3147043667317535705L;

	private final TableName tableName;

	protected TableDefModel defModel;

	protected String selectSql;

	protected List<Class<?>> classList;

	public TableDataModel(TableName tableName, TableDefModel defModel) {
		this.tableName = tableName;
		this.defModel = defModel;

		StringBuilder sb = new StringBuilder(512);

		int cols = defModel.getRowCount();
		classList = new ArrayList<Class<?>>(cols);

		for (int i = 0; i < cols; i++) {
			Object colName = defModel.getValueAt(i,
					TableDefModel.COLUMN_NAME_POS);

			if (i != 0) {
				sb.append(',');
			}
			sb.append(colName);
		}

		selectSql = "select " + sb + " from " + tableName.getFullTableName();

		refreshData(true);
	}

	public void refreshData(final boolean init) {
		setRowCount(0);

		new ExecuteQuery() {
			@Override
			protected void processResultSet(ResultSet rs) throws SQLException {
				if (init) {
					ResultSetMetaData md = rs.getMetaData();
					ToolUtil.addTableModelColumn(TableDataModel.this, md,
							classList);
				}
				super.processResultSet(rs);
			}

			@Override
			protected void setResultSet(ResultSet rs) throws SQLException {
				int cols = getColumnCount();
				Vector<Object> vec = new Vector<Object>(cols);
				for (int i = 1; i <= cols; i++) {
					Object v = rs.getObject(i);
					Data data = newData(i - 1, v);
					vec.add(data);
				}
				addRow(vec);
			}
		}.start(selectSql);
	}

	@Override
	public Data getDefaultValueAt(int row, int column) {
		return newData(column, null);
	}

	public Data newData(int mc, Object value) {
		Class<?> clz = getColumnClass(mc);
		Data data = new Data(clz);
		if (value != null) {
			data.setValue(value);
		}
		return data;
	}

	@Override
	public Class<?> getColumnClass(int columnIndex) {
		return classList.get(columnIndex);
	}

	@Override
	public Data getValueAt(int row, int column) {
		return (Data) super.getValueAt(row, column);
	}

	@SuppressWarnings("unchecked")
	@Override
	public void commit() {
		Connection conn;
		try {
			conn = ConnManager.getConnection();
		} catch (RuntimeException e) {
			DerbySchemaFrame.log(e);
			throw e;
		}
		try {

			loop: for (int i = getRowCount() - 1; i >= 0; i--) {
				Vector rowVector = (Vector) dataVector.elementAt(i);
				for (Object v : rowVector) {
					if (v instanceof ChangeValue) {
						ChangeValue cv = (ChangeValue) v;
						switch (cv.getType()) {
						case LazyTable.DEL:
							delete(conn, i);
							continue loop;
						}
					}
				}
			}
			loop: for (int i = 0; i < getRowCount(); i++) {
				Vector rowVector = (Vector) dataVector.elementAt(i);
				for (Object v : rowVector) {
					if (v instanceof ChangeValue) {
						ChangeValue cv = (ChangeValue) v;
						switch (cv.getType()) {
						case LazyTable.ADD:
							insert(conn, i);
							continue loop;
						case LazyTable.CHG:
							update(conn, i);
							continue loop;
						}
					}
				}
			}
			conn.commit();
		} catch (SQLException e) {
			DerbySchemaFrame.log(e);
			try {
				conn.rollback();
			} catch (SQLException e1) {
				DerbySchemaFrame.log(e1);
			} catch (RuntimeException e1) {
				DerbySchemaFrame.log(e1);
				throw e1;
			}
			return;
		}

		super.commit();
	}

	protected void insert(Connection conn, int row) throws SQLException {
		List<Object> vlist = new ArrayList<Object>();

		StringBuilder sb = new StringBuilder(512);
		sb.append("insert into ");
		sb.append(tableName.getFullTableName());
		sb.append(" (");
		for (int i = 0; i < getColumnCount(); i++) {
			String name = (String) defModel.getValueAt(i,
					TableDefModel.COLUMN_NAME_POS);
			if (i != 0) {
				sb.append(", ");
			}
			sb.append(name);
		}
		sb.append(") values (");
		for (int i = 0; i < getColumnCount(); i++) {
			if (i != 0) {
				sb.append(",");
			}
			Data data = getValueAt(row, i);
			if (data.isKeyword()) {
				sb.append(data.getValue());
			} else {
				sb.append("?");
				vlist.add(data.getValue());
			}
		}
		sb.append(")");
		String sql = sb.toString();

		PreparedStatement stat = conn.prepareStatement(sql);
		try {
			for (int i = 0; i < vlist.size(); i++) {
				Object v = vlist.get(i);
				DerbySchemaFrame.logln(String.format("%3d: %s", i + 1, v));
				stat.setObject(i + 1, v);
			}

			int r = stat.executeUpdate();
			DerbySchemaFrame.logln("insert: " + r);
		} finally {
			stat.close();
		}
	}

	protected void update(Connection conn, int row) throws SQLException {
		List<Object> vlist = new ArrayList<Object>();

		StringBuilder sb = new StringBuilder(512);
		sb.append("update ");
		sb.append(tableName.getFullTableName());
		sb.append(" set ");
		boolean first = true;
		for (int i = 0; i < getColumnCount(); i++) {
			if (getStatusAt(row, i) == LazyTable.CHG) {
				String name = (String) defModel.getValueAt(i,
						TableDefModel.COLUMN_NAME_POS);
				if (first) {
					first = false;
				} else {
					sb.append(", ");
				}
				sb.append(name);
				Data data = getValueAt(row, i);
				if (data.isKeyword()) {
					sb.append("=");
					sb.append(data.getValue());
				} else {
					sb.append("=?");
					vlist.add(data.getValue());
				}
			}
		}
		sb.append(" where ");
		appendWhereUniqueKey(conn, row, sb, vlist);

		String sql = sb.toString();
		PreparedStatement stat = conn.prepareStatement(sql);
		try {
			for (int i = 0; i < vlist.size(); i++) {
				Object v = vlist.get(i);
				DerbySchemaFrame.logln(String.format("%3d: %s", i + 1, v));
				stat.setObject(i + 1, v);
			}

			int r = stat.executeUpdate();
			DerbySchemaFrame.logln("update: " + r);
		} finally {
			stat.close();
		}
	}

	protected void delete(Connection conn, int row) throws SQLException {
		List<Object> vlist = new ArrayList<Object>();

		StringBuilder sb = new StringBuilder(512);
		sb.append("delete from ");
		sb.append(tableName.getFullTableName());
		sb.append(" where ");
		appendWhereUniqueKey(conn, row, sb, vlist);
		String sql = sb.toString();

		PreparedStatement stat = conn.prepareStatement(sql);
		try {
			for (int i = 0; i < vlist.size(); i++) {
				Object v = vlist.get(i);
				DerbySchemaFrame.logln(String.format("%3d: %s", i + 1, v));
				stat.setObject(i + 1, v);
			}

			int r = stat.executeUpdate();
			DerbySchemaFrame.logln("delete: " + r);
		} finally {
			stat.close();
		}
	}

	protected List<String> keyList;

	protected void appendWhereUniqueKey(Connection conn, int row,
			StringBuilder sb, List<Object> list) throws SQLException {
		if (keyList == null) {
			DatabaseMetaData md = conn.getMetaData();
			ResultSet rs = md.getBestRowIdentifier(null, tableName
					.getSchemaName(), tableName.getTableName(),
					DatabaseMetaData.bestRowTransaction, true);
			try {
				keyList = new ArrayList<String>();
				short scope = -1;
				while (rs.next()) {
					short s = rs.getShort("SCOPE");
					if (scope == -1) {
						scope = s;
					}
					if (s == scope) {
						keyList.add(rs.getString("COLUMN_NAME"));
					}
				}
			} finally {
				rs.close();
			}
		}

		for (int i = 0; i < keyList.size(); i++) {
			if (i != 0) {
				sb.append(" and ");
			}

			String name = keyList.get(i);
			sb.append(name);

			int n = getColumnIndex(name);
			Data data = (Data) getOldValueAt(row, n);
			Object value = data.getValue();
			if (value == null) {
				sb.append(" is null");
			} else {
				if (data.isKeyword()) {
					if ("null".equalsIgnoreCase((String) value)) {
						sb.append(" is ");
						sb.append(value);
					} else {
						sb.append("=");
						sb.append(value);
					}
				} else {
					sb.append("=?");
					list.add(value);
				}
			}

		}
	}

	public int getColumnIndex(String name) {
		for (int i = 0; i < defModel.getRowCount(); i++) {
			Object colName = defModel.getValueAt(i,
					TableDefModel.COLUMN_NAME_POS);
			if (name.equals(colName)) {
				return i;
			}
		}
		return -1;
	}
}